"""fast_simplification version

On the ``main`` branch, use 'dev0' to denote a development version.
For example:

version_info = 0, 27, 'dev0'

"""

version_info = 0, 1, 13
__version__ = ".".join(map(str, version_info))
